@@ -461,6 +461,8 @@
 #define PCI_CHIP_PALM_9803 0x9803
 #define PCI_CHIP_PALM_9804 0x9804
 #define PCI_CHIP_PALM_9805 0x9805
+#define PCI_CHIP_PALM_9806 0x9806
+#define PCI_CHIP_PALM_9807 0x9807
 #define PCI_CHIP_CYPRESS_6880 0x6880
 #define PCI_CHIP_CYPRESS_6888 0x6888
 #define PCI_CHIP_CYPRESS_6889 0x6889
@@ -469,6 +471,7 @@
 #define PCI_CHIP_CYPRESS_688D 0x688D
 #define PCI_CHIP_CYPRESS_6898 0x6898
 #define PCI_CHIP_CYPRESS_6899 0x6899
+#define PCI_CHIP_CYPRESS_689B 0x689B
 #define PCI_CHIP_CYPRESS_689E 0x689E
 #define PCI_CHIP_HEMLOCK_689C 0x689C
 #define PCI_CHIP_HEMLOCK_689D 0x689D
@@ -479,7 +482,9 @@
 #define PCI_CHIP_JUNIPER_68B0 0x68B0
 #define PCI_CHIP_JUNIPER_68B8 0x68B8
 #define PCI_CHIP_JUNIPER_68B9 0x68B9
+#define PCI_CHIP_JUNIPER_68BA 0x68BA
 #define PCI_CHIP_JUNIPER_68BE 0x68BE
+#define PCI_CHIP_JUNIPER_68BF 0x68BF
 #define PCI_CHIP_REDWOOD_68C0 0x68C0
 #define PCI_CHIP_REDWOOD_68C1 0x68C1
 #define PCI_CHIP_REDWOOD_68C7 0x68C7
@@ -514,6 +519,7 @@
 #define PCI_CHIP_CAYMAN_6719 0x6719
 #define PCI_CHIP_CAYMAN_671C 0x671C
 #define PCI_CHIP_CAYMAN_671D 0x671D
+#define PCI_CHIP_CAYMAN_671F 0x671F
 #define PCI_CHIP_BARTS_6720 0x6720
 #define PCI_CHIP_BARTS_6721 0x6721
 #define PCI_CHIP_BARTS_6722 0x6722
@@ -526,6 +532,7 @@
 #define PCI_CHIP_BARTS_6729 0x6729
 #define PCI_CHIP_BARTS_6738 0x6738
 #define PCI_CHIP_BARTS_6739 0x6739
+#define PCI_CHIP_BARTS_673E 0x673E
 #define PCI_CHIP_TURKS_6740 0x6740
 #define PCI_CHIP_TURKS_6741 0x6741
 #define PCI_CHIP_TURKS_6742 0x6742
